/*  CAMPUS LIsP, Lemon Version                                         */
/*  Copyright (C) 2000  Hirotsugu Kakugawa (h.kakugawa@computer.org)   */
/*  This is a free software. See "COPYING" for detail.                 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/* Data Representation ('int' must be at least 32 bits) */
typedef int    lobj;
#define D_MASK_DATA     0x0fffffff
#define D_MASK_TAG      0x70000000
#define D_GC_MARK       0x80000000
#define D_TAG_BIT_POS   28
#define D_INT_SIGN_BIT  0x08000000
#define D_GET_TAG(s)    (s & ~(D_GC_MARK | D_MASK_DATA))
#define D_GET_DATA(s)   (s & D_MASK_DATA)

/* Data Tags */
#define TAG_NIL     (0 << D_TAG_BIT_POS)
#define TAG_T       (1 << D_TAG_BIT_POS)
#define TAG_INT     (2 << D_TAG_BIT_POS)
#define TAG_SYMB    (3 << D_TAG_BIT_POS)
#define TAG_CONS    (4 << D_TAG_BIT_POS)
#define TAG_EOF     (5 << D_TAG_BIT_POS)
#define TAG_UNDEF   (6 << D_TAG_BIT_POS)

/* Cells */
#define NCONS   1024
static int   t_cons_free;           /* free list */
static lobj  t_cons_car[NCONS];     /* "car" part of cell */
static lobj  t_cons_cdr[NCONS];     /* "cdr" part of cell */

/* Symbols */
#define NSYMBS   64
static int   t_symb_free;           /* free slot */
static char *t_symb_pname[NSYMBS];  /* pointer to printable name */
static lobj  t_symb_val[NSYMBS];    /* symbol value */
static lobj  t_symb_fval[NSYMBS];   /* symbol function definition */
static int   t_symb_ftype[NSYMBS];  /* function type */ 

/* Printable name */
#define PNAME_SIZE   512
static int   t_pnames_free;         /* free pointer */
static char  t_pnames[PNAME_SIZE];  /* names */ 

/* Stack */
#define STACK_SIZE   128
static lobj  t_stack[STACK_SIZE];   /* the stack */
static int   t_stack_ptr;           /* stack pointer */

/* Function types */
enum Ftype {
  FTYPE_UNDEF,
  FTYPE_SYS, 
  FTYPE_SPECIAL,
  FTYPE_USER
};
#define FTYPE(t, nargs)       ((t)*1000 + (nargs))
#define FTYPE_ANY_ARGS         999
#define FTYPE_GET_TYPE(p)     ((p) / 1000)
#define FTYPE_GET_NARGS(p)    ((p) % 1000) 

/* Keywords */
enum keywords {
  KW_READ,    KW_EVAL,    KW_GC,      KW_CONS,    KW_CAR,      KW_CDR,
  KW_QUIT,    KW_DEFUN,   KW_LAMBDA,  KW_QUOTE,   KW_SETQ,     KW_EQ, 
  KW_NULL,    KW_ATOM,    KW_NUMBERP, KW_PRINT,   KW_TERPRI,   KW_RPLACA,
  KW_RPLACD,  KW_PROGN,   KW_COND,    KW_AND,     KW_OR,       KW_NOT,
  KW_LIST,    KW_ADD,     KW_SUB,     KW_TIMES,   KW_QUOTIENT, KW_DIVIDE,
  KW_GT,      KW_LT
};
struct s_keywords {
  char  *key;
  int    ftype;
  int    i;
};

/* Built-in function table */
static struct s_keywords funcs[] = {
  { "read",     FTYPE(FTYPE_SYS,     0),               KW_READ     },
  { "eval",     FTYPE(FTYPE_SYS,     1),               KW_EVAL     },
  { "gc",       FTYPE(FTYPE_SYS,     0),               KW_GC       },
  { "cons",     FTYPE(FTYPE_SYS,     2),               KW_CONS     },
  { "car",      FTYPE(FTYPE_SYS,     1),               KW_CAR      },
  { "cdr",      FTYPE(FTYPE_SYS,     1),               KW_CDR      },
  { "quit",     FTYPE(FTYPE_SYS,     0),               KW_QUIT     },
  { "defun",    FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_DEFUN    },
  { "lambda",   FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_LAMBDA   },
  { "quote",    FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_QUOTE    },
  { "setq",     FTYPE(FTYPE_SPECIAL, 2),               KW_SETQ     },
  { "eq",       FTYPE(FTYPE_SYS,     2),               KW_EQ       },
  { "null",     FTYPE(FTYPE_SYS,     1),               KW_NULL     },
  { "atom",     FTYPE(FTYPE_SYS,     1),               KW_ATOM     },
  { "numberp",  FTYPE(FTYPE_SYS,     1),               KW_NUMBERP  },
  { "print",    FTYPE(FTYPE_SYS,     1),               KW_PRINT    },
  { "terpri",   FTYPE(FTYPE_SYS,     0),               KW_TERPRI   },
  { "rplaca",   FTYPE(FTYPE_SYS,     2),               KW_RPLACA   },
  { "rplacd",   FTYPE(FTYPE_SYS,     2),               KW_RPLACD   },
  { "progn",    FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_PROGN    },
  { "cond",     FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_COND     },
  { "and",      FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_AND      },
  { "or",       FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_OR       },
  { "not",      FTYPE(FTYPE_SYS,     1),               KW_NOT      },
  { "list",     FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_LIST     },
  { "+",        FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_ADD      },
  { "-",        FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_SUB      },
  { "*",        FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_TIMES    },
  { "/",        FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_QUOTIENT },
  { "divide",   FTYPE(FTYPE_SYS,     2),               KW_DIVIDE   },
  { ">",        FTYPE(FTYPE_SYS,     2),               KW_GT       },
  { "<",        FTYPE(FTYPE_SYS,     2),               KW_LT       },
  { NULL,       -1,                                    -1          }
};

/* Error messages */
static char *errmsg_sym_undef  = "SYMBOL UNDEFINED: ";
static char *errmsg_func_undef = "FUNCTION UNDEFINED: ";
static char *errmsg_ill_nargs  = "ILLEGAL NUMBER OF ARGUMENTS: ";
static char *errmsg_ill_type   = "ILLEGAL ARGUMENT TYPE: ";
static char *errmsg_ill_call   = "ILLEGAL FUNCTION CALL: ";
static char *errmsg_ill_syntax = "ILLEGAL SYNTAX: ";
static char *errmsg_eof        = "END OF FILE";
static char *errmsg_stack_of   = "STACK OVERFLOW";
static char *errmsg_zero_div   = "DIVISION BY ZERO: ";

/* Function types */
static void   init(void), toplevel(void);
static lobj   l_read(void), l_eval(lobj s), l_print(lobj s);
static int    skip_space(void);
static lobj   int_make_l(int v);
static int    int_get_c(lobj s);
static lobj   eval_args(lobj func, lobj a, lobj av[2], int n);
static lobj   special(lobj f, lobj a);
static lobj   fcall(lobj f, lobj av[2], int n);
static lobj   apply(lobj f, lobj args, int n);
static int    err_msg(char *msg, int f, lobj s);
static lobj   l_cons(lobj car, lobj cdr), l_car(lobj s), l_cdr(lobj s);
static void   quit(void);
static int    list_len(lobj s);
static void   rplacd(lobj s, lobj cdr);
static void   gcollect(void);
static void   gc_mark(lobj s);
static int    gc_protect(lobj s);
static void   gc_unprotect(lobj s);
static lobj   symb_make(char *p);


int
main(int argc, char **argv)
{
  init();
  printf("CAMPUS LIsP, Lemon version 1.1\n");
  toplevel();
  quit();

  return 0;
}


/* Initialize lisp strage */
static void
init(void)
{
  int   i;

  /* cells */
  t_cons_free = 0;

  /* make a free cell list */
  for (i = 0; i < NCONS-1; i++)
    t_cons_car[i]  = i+1;
  t_cons_car[NCONS-1] = NCONS-1;    /* self-loop (this is very important) */

  /* symbol table */
  t_symb_free = 0;
  for (i = 0; i < NSYMBS; i++)
    t_symb_pname[i] = NULL;

  /* print name (symbol name) */
  t_pnames_free = 0;

  /* stack */
  t_stack_ptr = 0;

  /* install built-in functions */
  for (i = 0; funcs[i].key != NULL; i++){
    symb_make(funcs[i].key);
    t_symb_ftype[i] = funcs[i].ftype;
    if (i != funcs[i].i){
      printf("function install error: %s\n", funcs[i].key);
      quit();
    }
  }
}


/* Top level */
static void
toplevel(void)
{
  lobj  s, v;

  for (;;){
    t_stack_ptr = 0;
    printf("\n] ");             /* prompt */
    if ((s = l_read()) < 0)     /* read */
      continue;
    if (s == TAG_EOF)           /* end of file */
      break;
    if (gc_protect(s) < 0)
      break;
    if ((v = l_eval(s)) < 0)    /* eval */
      continue;
    gc_unprotect(s);
    (void) l_print(v);          /* print */
  }
}


/* Read an S-expression */
static lobj
l_read(void)
{
  lobj  s, v, t;
  char  token[32];
  int   ch, i;
  
  /* skip spaces */
  if ((ch = skip_space()) < 0){  /* eof */
    return TAG_EOF; 

  } else if (ch == ';'){         /* comment */
    while (getc(stdin) != '\n')
      ;
    
  } else if (ch == '\''){        /* quote macro */
    if ((t = l_read()) < 0)
      return -1;
    if (t == TAG_EOF)
      return err_msg(errmsg_eof, 0, 0);
    t = l_cons(t, TAG_NIL);
    s = l_cons((TAG_SYMB|KW_QUOTE), t);

  } else if (ch != '('){         /* t, nil, symbol, or integer */
    token[0] = ch;
    for (i = 1; ; i++){
      ch = getc(stdin);
      if (isspace(ch) || iscntrl(ch) || (ch < 0) 
	  || (ch == ';') || (ch == '(') || (ch == ')')){
	ungetc(ch, stdin);
	token[i] = '\0';
	if (isdigit((int)token[0]) 
	    || ((token[0] == '-') && isdigit((int)token[1]))
	    || ((token[0] == '+') && isdigit((int)token[1]))){   /* integer */
	  s = int_make_l(atoi(token));
	} else if (strcmp(token, "nil") == 0){                   /* nil */ 
	  s = TAG_NIL;
	} else if (strcmp(token, "t") == 0){                     /* t */
	  s = TAG_T;
	} else {                                                 /* symbol */
	  s = TAG_SYMB | symb_make(token);
	}
	break;
      }
      token[i] = ch;
    }

  } else /* ch == '(' */ {       /* list */
    if ((ch = skip_space()) < 0){
      return err_msg(errmsg_eof, 0, 0);
    } else if (ch == ')'){
      s = TAG_NIL;  /* "()" = nil */
    } else {
      ungetc(ch, stdin);
      if ((t = l_read()) < 0)
	return err_msg(errmsg_eof, 0, 0);
      if (t == TAG_EOF)
	return -1;
      if ((s = v = l_cons(t, TAG_NIL)) < 0)
	return -1;
      if (gc_protect(s) < 0)
	return -1;
      for (;;){
	if ((ch = skip_space()) < 0)  /* look ahead next char */
	  return err_msg(errmsg_eof, 0, 0);
	if (ch == ')')
	  break;
	ungetc(ch, stdin);
	if ((t = l_read()) < 0)
	  return -1;
	if (t == TAG_EOF)
	  return err_msg(errmsg_eof, 0, 0);
	if ((t = l_cons(t, TAG_NIL)) < 0) 
	  return -1;
	rplacd(v, t);
	v = l_cdr(v);
      }
      gc_unprotect(s);
    }
  }

  return s;
}

static int
skip_space(void)
{
  int  ch;

  for (;;){
    if ((ch = getc(stdin)) < 0)
      return -1;     /* end-of-file */
    if (!isspace(ch) && !iscntrl(ch))
      break;
  }
  return ch;
}


/* Print an S-expression */ 
static lobj
l_print(lobj s)
{
  lobj  v, t;
  int   i;

  switch(D_GET_TAG(s)){

  case TAG_NIL:
    printf("nil");
    break;

  case TAG_T:
    printf("t");
    break;

  case TAG_INT:
    v = int_get_c(s);
    printf("%d", v);
    break;

  case TAG_SYMB:
    i = s & D_MASK_DATA;
    printf("%s", t_symb_pname[i]);
    break;

  case TAG_EOF:
    printf("<eof>");
    break;

  case TAG_UNDEF:  /* for debugging */
    printf("<undefined>"); 
    break;

  case TAG_CONS:
    printf("(");
    t = s;
    l_print(l_car(t));
    while (D_GET_TAG(l_cdr(t)) == TAG_CONS) {
      printf(" ");
      t = l_cdr(t);
      l_print(l_car(t));
    }
    if (D_GET_TAG(l_cdr(t)) != TAG_NIL){
      printf(" . ");
      l_print(l_cdr(t)); 
    }
    printf(")");
    break;
  }
  return TAG_T;
}


/* Evaluate an S-expression */
static lobj
l_eval(lobj s)
{
  lobj  v, f, a, av[2];
  int   n; 

  switch(D_GET_TAG(s)){

  case TAG_NIL:        /* self-evaluating objects */
  case TAG_T:
  case TAG_INT:
    v = s;
    break;

  case TAG_SYMB:       /* symbol ... refer to the symbol table */
    if ((v = t_symb_val[D_GET_DATA(s)]) == TAG_UNDEF)
      return err_msg(errmsg_sym_undef, 1, s);
    break;

  case TAG_CONS:       /* cons ... function call */
    f = l_car(s);   /* function name or lambda exp */
    a = l_cdr(s);   /* actual argument list */
    if ((D_GET_TAG(f) == TAG_CONS) && (D_GET_TAG(l_car(f)) == TAG_SYMB) 
	&& (D_GET_DATA(l_car(f)) == KW_LAMBDA)){   /* lambda exp */
      if (eval_args(f, a, av, FTYPE_ANY_ARGS) < 0)
	return -1;
      v = apply(l_cdr(f), av[0], list_len(l_car(l_cdr(f))));
    } else if (D_GET_TAG(f) == TAG_SYMB){
      n = FTYPE_GET_NARGS(t_symb_ftype[D_GET_DATA(f)]);
      switch (FTYPE_GET_TYPE(t_symb_ftype[D_GET_DATA(f)])){
      case FTYPE_UNDEF:
	return err_msg(errmsg_func_undef, 1, f);
      case FTYPE_SPECIAL:
	v = special(f, a);
	break;
      case FTYPE_SYS:
	if (eval_args(f, a, av, n) < 0)
	  return -1;
	v = fcall(f, av, n);
	break;
      case FTYPE_USER:
	if (eval_args(f, a, av, FTYPE_ANY_ARGS) < 0)
	  return -1;
	v = apply(f, av[0], n);
      }
    } else {
      return err_msg(errmsg_ill_call, 1, s);
    }
    break;
  }
  return v;
}


/* Execute special form (defun, setq. etc... arguments are not evaluated) */
static lobj
special(lobj f, lobj a)
{
  lobj  t, v, u;
  int   l, i;

  switch (D_GET_DATA(f)){

  case KW_DEFUN:
    if (list_len(a) < 2)
      return err_msg(errmsg_ill_syntax, 1, f);
    v = l_car(a);            /* function name  */
    if (D_GET_TAG(v) != TAG_SYMB)
      return err_msg(errmsg_ill_syntax, 1, f);
    t = l_cdr(a);            /* list of function arg and function body */
    l = list_len(l_car(t));  /* #args */
    i = D_GET_DATA(v);
    t_symb_fval[i]  = t;
    t_symb_ftype[i] = FTYPE(FTYPE_USER, l);
    break;

  case KW_SETQ:
    t = l_car(a);  /* symbol name */
    if (D_GET_TAG(t) != TAG_SYMB)
      return err_msg(errmsg_ill_type, 1, f);
    if ((v = l_eval(l_car(l_cdr(a)))) < 0)  /* value */
      return -1;
    t_symb_val[D_GET_DATA(t)] = v;
    break;

  case KW_QUOTE:
    v = l_car(a);
    break;

  case KW_PROGN:
    for (v = TAG_NIL, t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
      if ((v = l_eval(l_car(t))) < 0)
	return -1;
    }
    break;

  case KW_AND:
    for (v = TAG_T, t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
      if ((v = l_eval(l_car(t))) < 0)
	return -1;
      if (D_GET_TAG(v) == TAG_NIL)
	break;
    }
    break;

  case KW_OR:
    for (v = TAG_NIL, t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
      if ((v = l_eval(l_car(t))) < 0)
	return -1;
      if (D_GET_TAG(v) != TAG_NIL)
	break;
    }
    break; 

  case KW_COND:
    if (D_GET_TAG(a) != TAG_CONS)
      return err_msg(errmsg_ill_syntax, 1, f);
    v = TAG_NIL; 
    for (t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
      u = l_car(t);
      if (D_GET_TAG(u) != TAG_CONS)
	return err_msg(errmsg_ill_syntax, 1, f);
      if ((v = l_eval(l_car(u))) < 0)
	return -1;
      if (D_GET_TAG(v) != TAG_NIL){
	for (u = l_cdr(u); D_GET_TAG(u) == TAG_CONS; u = l_cdr(u)){ 
	  if ((v = l_eval(l_car(u))) < 0)
	    return -1;
	} 
	break;
      }
    }
    break;
  }
  return v;
}


/* Evaluate arguments */
static lobj
eval_args(lobj func, lobj arg, lobj av[2], int n)
{
  lobj  x, y;

  if ((n != FTYPE_ANY_ARGS) && (n != list_len(arg)))
    return err_msg(errmsg_ill_nargs, 1, func);

  switch (n){

  case 0:
    av[0] = TAG_NIL;
    break;

  case 1:
    if ((av[0] = l_eval(l_car(arg))) < 0)
      return -1;
    break;

  case 2:
    if ((av[0] = l_eval(l_car(arg))) < 0)
      return -1;
    if (gc_protect(av[0]) < 0)
      return -1;
    if ((av[1] = l_eval(l_car(l_cdr(arg)))) < 0)
      return -1;
    gc_unprotect(av[0]);
    break;

  case FTYPE_ANY_ARGS:   /* return evaluated arguments as a list */
    if (D_GET_TAG(arg) != TAG_CONS){
      av[0] = TAG_NIL;
    } else {
      if ((x = l_eval(l_car(arg))) < 0)
	return -1;
      if ((av[0] = y = l_cons(x, TAG_NIL)) < 0)
	return -1;
      if (gc_protect(av[0]) < 0)
	return -1;
      for (arg = l_cdr(arg); D_GET_TAG(arg) == TAG_CONS; arg = l_cdr(arg)){
	if ((x = l_eval(l_car(arg))) < 0)
	  return -1;
	rplacd(y, l_cons(x, TAG_NIL)); 
	y = l_cdr(y);
      }
      gc_unprotect(av[0]);
    }
  }
  return av[0];
}


/* Call a built-in function */
static lobj 
fcall(lobj f, lobj av[2], int n)
{
  lobj  v, t;
  int   r, d;

  switch (D_GET_DATA(f)){

  case KW_LAMBDA:
    return err_msg(errmsg_ill_call, 1, f);

  case KW_QUIT:
    quit();
    break;
    
  case KW_EQ:
    v = (av[0] == av[1]) ? TAG_T : TAG_NIL;
    break;

  case KW_CONS:
    v = l_cons(av[0], av[1]); 
    break;

  case KW_RPLACA:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(errmsg_ill_type, 1, f);
    v = t_cons_car[D_GET_DATA(av[0])] = av[1];
    break;

  case KW_RPLACD:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(errmsg_ill_type, 1, f);
    v = t_cons_cdr[D_GET_DATA(av[0])] = av[1];
    break;

  case KW_CAR:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(errmsg_ill_type, 1, f);
    v = l_car(av[0]);
    break;

  case KW_CDR:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(errmsg_ill_type, 1, f);
    v = l_cdr(av[0]);
    break;

  case KW_NULL:
    v = (D_GET_TAG(av[0]) == TAG_NIL) ? TAG_T : TAG_NIL;
    break;

  case KW_ATOM:
    v = (D_GET_TAG(av[0]) != TAG_CONS) ? TAG_T : TAG_NIL;
    break;

  case KW_NUMBERP:
    v = (D_GET_TAG(av[0]) == TAG_INT) ? TAG_T : TAG_NIL;
    break;

  case KW_LIST:
    v = av[0];
    break;

  case KW_NOT:
    v = (D_GET_TAG(av[0]) == TAG_NIL) ? TAG_T : TAG_NIL;
    break;

  case KW_READ:
    v = l_read();
    break;

  case KW_EVAL:
    v = l_eval(av[0]);
    break;

  case KW_PRINT:
    v = l_print(av[0]);
    break;

  case KW_TERPRI:
    printf("\n");
    v = TAG_T;
    break;

  case KW_GC:
    gcollect();
    v = TAG_T;
    break;

  case KW_ADD:
    for (r = 0, t = av[0]; D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
      if (D_GET_TAG(l_car(t)) != TAG_INT)
	return err_msg(errmsg_ill_type, 1, f);
      r = r + int_get_c(l_car(t));
    }
    v = int_make_l(r);
    break;

  case KW_TIMES:
    for (r = 1, t = av[0]; D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
      if (D_GET_TAG(l_car(t)) != TAG_INT)
	return err_msg(errmsg_ill_type, 1, f);
      r = r * int_get_c(l_car(t));
    }
    v = int_make_l(r);
    break;

  case KW_SUB:
    if (D_GET_TAG(av[0]) == TAG_NIL){
      r = 0;
    } else if (D_GET_TAG(l_car(av[0])) != TAG_INT){
	return err_msg(errmsg_ill_type, 1, f);
    } else if (D_GET_TAG(l_cdr(av[0])) == TAG_NIL){
      r = 0 - int_get_c(l_car(av[0]));
    } else {
      r = int_get_c(l_car(av[0]));
      for (t = l_cdr(av[0]); D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
	if (D_GET_TAG(l_car(t)) != TAG_INT)
	  return err_msg(errmsg_ill_type, 1, f);
	r = r - int_get_c(l_car(t));
      }
    }
    v = int_make_l(r);
    break;

  case KW_QUOTIENT:
    if (D_GET_TAG(av[0]) == TAG_NIL){
      r = 1;
    } else if (D_GET_TAG(l_car(av[0])) != TAG_INT){
	return err_msg(errmsg_ill_type, 1, f);
    } else if ((d = int_get_c(l_car(av[0]))) == 0){
      return err_msg(errmsg_zero_div, 1, f);
    } if (D_GET_TAG(l_cdr(av[0])) == TAG_NIL){
      r = 1 / d;
    } else {
      for (r = d, t = l_cdr(av[0]); D_GET_TAG(t) == TAG_CONS; t = l_cdr(t)){
	if (D_GET_TAG(l_car(t)) != TAG_INT)
	  return err_msg(errmsg_ill_type, 1, f);
	if ((d = int_get_c(l_car(t))) == 0)
	  return err_msg(errmsg_zero_div, 1, f);
	r = r / d;
      }
    }
    v = int_make_l(r);
    break;

  case KW_DIVIDE:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(errmsg_ill_type, 1, f);
    r = int_get_c(av[0]);
    if ((d = int_get_c(av[1])) == 0)
      return err_msg(errmsg_zero_div, 1, f);
    v = l_cons(int_make_l(r / d), int_make_l(r % d));
    break;

  case KW_GT:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(errmsg_ill_type, 1, f);
    v = (int_get_c(av[0]) > int_get_c(av[1])) ? TAG_T : TAG_NIL;
    break;

  case KW_LT:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(errmsg_ill_type, 1, f);
    v = (int_get_c(av[0]) < int_get_c(av[1])) ? TAG_T : TAG_NIL;
    break;
  }
  return v;
}


/* Function application (user defined function) */
static lobj
apply(lobj func, lobj aparams, int n) 
{
  lobj  fdef, fbody, f, sym, a, v;
  int   i;

  if (t_stack_ptr + n + 1 >= STACK_SIZE)   /* stack overflow */
    return err_msg(errmsg_stack_of, 0, 0);

  if (D_GET_TAG(func) == TAG_SYMB){         /* function symbol */
    fdef = t_symb_fval[D_GET_DATA(func)];
  } else if (D_GET_TAG(func) == TAG_CONS){  /* lambda exp */
    fdef = func;
  }

  /* bind */
  f = l_car(fdef);  /* formal parameters */
  a = aparams;      /* actual parameters */
  t_stack_ptr = t_stack_ptr + n;
  for (i = 0; i < n; i++, f = l_cdr(f), a = l_cdr(a)){
    sym = l_car(f);
    /* push old symbol values onto stack */
    t_stack[t_stack_ptr - i - 1] = t_symb_val[D_GET_DATA(sym)];
    /* bind argument value to symbol */
    t_symb_val[D_GET_DATA(sym)] = l_car(a);
  }

  if (gc_protect(aparams) < 0)
    return -1;

  /* evaluate function body */
  fbody = l_cdr(fdef);  /* function body */
  for (v = TAG_NIL; D_GET_TAG(fbody) == TAG_CONS; fbody = l_cdr(fbody)){
    if ((v = l_eval(l_car(fbody))) < 0)
      break;   /* error ... never return immediately - need unbinding. */
  }

  /* pop gc_protected objects, including 'gc_unprotect(aparams)'. */
  while ((t_stack[t_stack_ptr-1] & D_GC_MARK) != 0)
    --t_stack_ptr;   

  /* unbind: restore old variable values from stack */
  for (i = 0, f = l_car(fdef); i < n; i++, f = l_cdr(f)){
    sym = l_car(f);
    t_symb_val[D_GET_DATA(sym)] = t_stack[t_stack_ptr - i - 1];
  }
  t_stack_ptr = t_stack_ptr - n;

  return v;
}


/* Print an error message */
static int
err_msg(char *msg, int f, lobj s)
{
  printf("\nERROR. \n%s", msg);
  if (f != 0)
    l_print(s);
  printf("\n");
  return -1;
}

/* Length of a list */
static int
list_len(lobj s)
{
  int  i;

  for (i = 0; D_GET_TAG(s) == TAG_CONS; s = l_cdr(s))
    i++;
  return i;
}

/* "Replace cdr" opration ... rewite cdr part of a cons cell */
static void
rplacd(lobj s, lobj cdr)
{
  t_cons_cdr[D_GET_DATA(s)] = cdr;
}


/* "Cons" opration */
static lobj
l_cons(lobj car, lobj cdr)
{
  int  s;

  if (t_cons_free < 0){   /*  no cons cells */
    if (gc_protect(car) < 0)
      return -1;
    if (gc_protect(cdr) < 0)
      return -1;
    gcollect();           /* invoke garbage collector */
    gc_unprotect(cdr);
    gc_unprotect(car);
  }

  /* get a free cons cell from a free list */
  s = t_cons_free;
  if (t_cons_car[t_cons_free] != t_cons_free)
    t_cons_free  = t_cons_car[t_cons_free];  /* next free cell */
  else
    t_cons_free = -1;                        /* self-loop: end of free list */

  /* constract a new cell */
  t_cons_car[s] = car;
  t_cons_cdr[s] = cdr;

  return (TAG_CONS | s);
}

/* "Car" opration */
static lobj
l_car(lobj s)
{
  return t_cons_car[D_GET_DATA(s)];
}

/* "Cdr" opration */
static lobj
l_cdr(lobj s)
{
  return t_cons_cdr[D_GET_DATA(s)];
}

/* Quit micro lisp */
static void
quit(void)
{
  printf("\nBYE\n");
  exit(0);
}


/* Garbage collector */
static void
gcollect(void)
{
  int  i, n, p;

  /* mark */
  for (i = 0; i < t_symb_free; i++)
    gc_mark(t_symb_val[i]);
  for (i = 0; i < t_symb_free; i++)
    gc_mark(t_symb_fval[i]);
  for (i = 0; i < t_stack_ptr; i++)
    gc_mark(t_stack[i]);

  /* sweep */
  t_cons_free = -1;
  for (i = 0, n = 0; i < NCONS; i++){
    if ((t_cons_car[i] & D_GC_MARK) == 0){  /* collect */
      n++;
      if (t_cons_free == -1){
	t_cons_free = i;
      } else {
	t_cons_car[p] = i; 
      }
      t_cons_car[i] = i; 
      p = i;
    }
    t_cons_car[i] &= ~D_GC_MARK;   /* clear mark */
  }

  if (n == 0){    /* no more cells... */
    printf("\nno memory. abort.\n");
    quit();
  }
}

/* mark recursively */
static void
gc_mark(lobj s)
{
  for ( ; D_GET_TAG(s) == TAG_CONS; s = l_cdr(s)){
    if ((t_cons_car[D_GET_DATA(s)] & D_GC_MARK) != 0) /* visited before */
      return;
    t_cons_car[D_GET_DATA(s)] |= D_GC_MARK;  /* mark */
    gc_mark(l_car(s));                       /* visit car part */
  }
}

/* protect/unprotect temporary objects from garbage collector */
static int
gc_protect(lobj s)
{
  if (D_GET_TAG(s) == TAG_CONS){  /* save only cons cells */
    if (t_stack_ptr >= STACK_SIZE)     /* stack overflow */
      return err_msg(errmsg_stack_of, 0, 0);
    t_stack[t_stack_ptr++] = (D_GC_MARK | s);  
  }
  return 0;
}

static void
gc_unprotect(lobj s)
{
  if (D_GET_TAG(s) == TAG_CONS)
    t_stack_ptr = t_stack_ptr - 1;
}


/* Make a Lisp integer from a C integer */
static lobj 
int_make_l(int v)
{
  return (TAG_INT | ((unsigned int)v & D_MASK_DATA));
}

/* Make a C integer from a Lisp integer */
static int 
int_get_c(lobj s)
{
  if (((unsigned int)s & D_INT_SIGN_BIT) == 0)
    return ((unsigned int)s & D_MASK_DATA);
  return (int) ((unsigned int)s | ~D_MASK_DATA);
}


/* Make a new symbol */
static lobj
symb_make(char *p)
{
  int    s, i;

  for (i = 0; i < NSYMBS-1; i++){
    if ((t_symb_pname[i] != NULL) && (strcmp(p, t_symb_pname[i]) == 0))
      return i;
  }

  s = t_symb_free;
  t_symb_free++;
  if (t_symb_free == NSYMBS){
    printf("\nno memory. abort.\n");
    quit();
  }

  t_symb_pname[s] = &t_pnames[t_pnames_free];
  t_symb_val[s]   = TAG_UNDEF;   /* undefined value */
  t_symb_fval[s]  = TAG_UNDEF;   /* undefined value */
  t_symb_ftype[s] = FTYPE_UNDEF; /* undefined type */

  do {
    if (t_pnames_free == PNAME_SIZE){
      printf("\nno memory. abort.\n");
      quit();
    }
    t_pnames[t_pnames_free++] = tolower(*(p++));
  } while (*p != '\0');
  t_pnames[t_pnames_free++] = '\0';

  return (TAG_SYMB | s);
}

/* END */
